import os
from pathlib import Path

from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, HTMLResponse
from fastapi.staticfiles import StaticFiles

from .routes import router

environment = os.getenv("ENV", "prod")
react_app_port = os.getenv("REACT_APP_PORT", "3000")

app = FastAPI()

if environment != "prod":
    origins = [
        "http://localhost:" . react_app_port,
        "http://127.0.0.1:" . react_app_port,
    ]

    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

app.include_router(router, prefix="/api")

BASE_DIR = Path(__file__).resolve().parent.parent
static_files_path = BASE_DIR / "static" / "static"

if environment == "prod":
    app.mount(
        "/static",
        StaticFiles(
            directory=str(static_files_path)),
        name="static")

    static_index_path = BASE_DIR / "static" / "index.html"

    @app.get("/{full_path:path}", response_class=HTMLResponse)
    async def catch_all(request: Request, full_path: str):
        if ".." in full_path:
            return HTMLResponse("Access Denied", status_code=403)

        file_path = BASE_DIR / "static" / full_path

        if file_path.suffix in ['.png', '.json', '.ico', '.js']:
            if file_path.is_file():
                return FileResponse(file_path)
            else:
                return HTMLResponse("File not found", status_code=404)

        if not request.url.path.startswith("/api"):
            return static_index_path.read_text()

        return HTMLResponse("Not found", status_code=404)



from datetime import datetime, timezone
from functools import partial
from typing import Dict, List, NamedTuple, Optional

from ...domain.model.completion_log import CompletionLog
from ...domain.model.periodicity import Periodicity


class StreakData(NamedTuple):
    """
    Information about a streaks.

    Attributes:
        current_streak (int): Current streak length.
        max_streak (int): Maximum streak length.
        broken_streaks (int): Number of broken streaks.
        consecutive_streaks (Dict[datetime, int]): Consecutive streaks.
    """
    current_streak: int
    max_streak: int
    broken_streaks: int
    consecutive_streaks: Dict[datetime, int]


class Monad:
    """
    A simple implementation of a monad.

    Attributes:
        value: The value contained within the monad.
    """

    def __init__(self, value):
        self.value = value

    def bind(self, func):
        """Applies a function to the value within the monad."""
        return Monad(func(self.value))


def is_log_within_period(
        log_date: datetime,
        period_start: datetime,
        period_end: datetime) -> bool:
    """
    Checks if a log date is within a specified period.

    Args:
        log_date (datetime): The log date.
        period_start (datetime): The start date of the period.
        period_end (datetime): The end date of the period.

    Returns:
        bool: True if log_date is within the period, False otherwise.
    """
    return period_start < log_date <= period_end


def update_streak_data(
        log_date: datetime,
        period_start: datetime,
        period_end: datetime,
        today: datetime,
        streak_data: StreakData) -> StreakData:
    """
    Updates streak data based on a log date.

    Args:
        log_date (datetime): The log date.
        period_start (datetime): The start date of the streak period.
        period_end (datetime): The end date of the streak period.
        today (datetime): The current date.
        streak_data (StreakData): The current streak data.

    Returns:
        StreakData: Updated streak data.
    """
    current_streak, max_streak, broken_streaks, consecutive_streaks = streak_data
    if is_log_within_period(log_date, period_start, period_end):
        current_streak += 1
        consecutive_streaks = {**consecutive_streaks, log_date: current_streak}
    elif period_end <= today and current_streak > 0:
        broken_streaks += 1
        current_streak = 0
        consecutive_streaks = {
            **consecutive_streaks, period_end: current_streak}

    max_streak = max(current_streak, max_streak)
    return StreakData(
        current_streak,
        max_streak,
        broken_streaks,
        consecutive_streaks)


def calculate_streak(
        log_dates: List[CompletionLog],
        period_start: datetime,
        period_end: datetime,
        streak_data: StreakData,
        today: datetime) -> Monad:
    """
    Calculates the streak data for a given period.

    Args:
        log_dates (List[CompletionLog]): List of completion logs.
        period_start (datetime): The start date of the period.
        period_end (datetime): The end date of the period.
        streak_data (StreakData): The current streak data.
        today (datetime): The current date.

    Returns:
        Monad: Monad containing the updated streak data.
    """
    process_log_for_streak = partial(
        update_streak_data,
        period_start=period_start,
        period_end=period_end,
        today=today,
        streak_data=streak_data)
    streaks = (process_log_for_streak(log_date) for log_date in log_dates)
    max_streak_data = max(
        streaks, key=lambda streak: streak.current_streak, default=streak_data)
    return Monad(max_streak_data)


def update_streaks_data_monad(
        current_date: datetime,
        streak_data_monad: Monad,
        log_dates: List[CompletionLog],
        periodicity: Periodicity,
        today: datetime) -> Monad:
    """
    Recursively updates streak data for a sequence of periods.

    Args:
        current_date (datetime): The current date in the sequence.
        streak_data_monad (Monad): Monad containing the current streak data.
        log_dates (List[CompletionLog]): List of completion logs.
        periodicity: The periodicity.
        today (datetime): The current date.

    Returns:
        Monad: Monad containing the updated streak data.
    """
    if current_date > today:
        return streak_data_monad

    next_date = periodicity.get_next_date(current_date)
    updated_streak_data_monad = calculate_streak(
        log_dates, current_date, next_date, streak_data_monad.value, today)

    return update_streaks_data_monad(
        next_date,
        updated_streak_data_monad,
        log_dates,
        periodicity,
        today)


def get_streaks_info(
        logs: List[CompletionLog],
        periodicity: Periodicity,
        creation_date: datetime,
        today: Optional[datetime] = None) -> StreakData:
    """
    Computes the streak data for a given set of logs.

    Args:
        logs (List[CompletionLog]): List of completion logs.
        periodicity: The periodicity.
        creation_date (datetime): The date when the streak tracking started.
        today (Optional[datetime]): The current date.

    Returns:
        StreakData: The computed streak data.
    """
    if not logs:
        return StreakData(0, 0, 0, {})

    today = today or datetime.now(timezone.utc).replace(tzinfo=None)
    log_dates = sorted(log.date for log in logs)

    streaks_data_monad = update_streaks_data_monad(creation_date, Monad(
        StreakData(0, 0, 0, {})), log_dates, periodicity, today)

    return streaks_data_monad.value



from typing import Optional

from fastapi import APIRouter, Depends, Query
from starlette import status
from fastapi import HTTPException

from ..application.service.analytics_service import AnalyticsService
from ..application.service.habit_service import HabitService
from ..domain.model.periodicity import Periodicity
from .dependencies import get_analytics_service, get_habit_service
from .schemas import HabitCreateSchema

router = APIRouter()


@router.post("/habits/",
             status_code=201,
             operation_id="createHabit")
def create_habit(
        habit: HabitCreateSchema,
        habit_service: HabitService = Depends(get_habit_service)):
    try:
        return habit_service.create_habit(habit.description, habit.periodicity)
    except ValueError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))


@router.get("/habits/", operation_id="getAllHabits")
def list_habits(
    periodicity: Optional[str] = Query(None),
    habit_service: HabitService = Depends(get_habit_service),
    analytics_service: AnalyticsService = Depends(get_analytics_service)
):
    try:
        if periodicity:
            periodicity_enum = Periodicity(periodicity)
            return analytics_service.list_habits_by_periodicity(
                periodicity_enum)
        return habit_service.list_all_tracked_habits()
    except ValueError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))


@router.put("/habits/{habit_id}/complete", operation_id="completeHabitTask")
def complete_habit(
        habit_id: int,
        habit_service: HabitService = Depends(get_habit_service)):
    try:
        habit_service.complete_habit(habit_id)
        return {"status": "completed"}
    except ValueError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Habit not found")


@router.delete("/habits/{habit_id}", operation_id="deleteHabit")
def delete_habit(
        habit_id: int,
        habit_service: HabitService = Depends(get_habit_service)):
    try:
        habit_service.delete_habit(habit_id)
        return {"status": "deleted"}
    except ValueError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Habit not found")


@router.get("/habits/analytics/longest-streak",
            operation_id="getLongestStreakForAllHabits")
def get_longest_streak(
        analytics_service: AnalyticsService = Depends(get_analytics_service)):
    return {"streak": analytics_service.longest_streak_for_all_habits()}


@router.get("/habits/{habit_id}/analytics/streak",
            operation_id="getLongestStreakForHabit")
def get_longest_streak_for_habit(
        habit_id: int,
        analytics_service: AnalyticsService = Depends(get_analytics_service)):
    return {"streak": analytics_service.longest_streak_for_habit(habit_id)}


@router.get("/habits/analytics/struggled-last-month",
            operation_id="getStruggledHabitsLastMonth")
def get_struggled_habits_last_month(
        analytics_service: AnalyticsService = Depends(get_analytics_service)):
    return analytics_service.struggled_habits_last_month()


@router.get("/habits/{habit_id}", operation_id="getHabitDetails")
def get_habit(
        habit_id: int,
        habit_service: HabitService = Depends(get_habit_service)):
    habit = habit_service.get_habit_by_id(habit_id)
    if habit is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Habit not found")
    return habit



from datetime import datetime, timezone

from ...domain.model.habit import Habit
from ...domain.model.periodicity import Periodicity
from ...infrastructure.repository.habit_repository import HabitRepository
from .analytics_service import AnalyticsService


class HabitService:
    def __init__(self, repository: HabitRepository,
                 analytics_service: AnalyticsService):
        self.repository = repository
        self.analytics_service = analytics_service

    def create_habit(
            self,
            description: str,
            periodicity: Periodicity) -> Habit:
        habit = Habit(description=description, periodicity=periodicity)
        self.repository.add(habit)
        return habit

    def complete_habit(self, habit_id):
        habit = self.repository.find_by_id(habit_id)
        if not habit:
            raise ValueError(f"No habit found with id {habit_id}")
        habit.complete()
        self.repository.save(habit)

    def delete_habit(self, habit_id: int) -> None:
        habit = self.repository.find_by_id(habit_id)
        if not habit:
            raise ValueError(f"No habit found with id {habit_id}")
        if habit:
            self.repository.remove(habit)

    def get_habit_by_id(self, habit_id: int) -> dict:
        habit = self.repository.find_by_id(habit_id)
        if habit:
            analytics_data = self.get_analytics_data(habit.id)
            return {'habit': {
                "id": habit.id,
                "description": habit.description,
                "periodicity": habit.periodicity,
                "creation_date": habit.creation_date
            }, **analytics_data}
        return None

    def list_all_tracked_habits(self) -> list:
        habits = self.repository.get_all()
        result = []
        for habit in habits:
            analytics_data = self.get_analytics_data(habit.id)
            result.append(
                {
                    'habit': {
                        "id": habit.id,
                        "description": habit.description,
                        "periodicity": habit.periodicity,
                        "creation_date": habit.creation_date},
                    'is_current_period_marked': analytics_data['is_current_period_marked'],
                    'current_streak': analytics_data['current_streak']})
        return result

    def get_analytics_data(self, habit_id):
        habit = self.repository.find_by_id(habit_id)
        return {
            'is_current_period_marked': self.analytics_service.is_current_period_marked(
                habit.completion_logs,
                habit.periodicity,
                datetime.now(timezone.utc).replace(tzinfo=None)),
            'current_streak': self.analytics_service.current_streak(
                habit.completion_logs,
                habit.periodicity,
                habit.creation_date),
            'longest_streak': self.analytics_service.longest_streak(
                habit.completion_logs,
                habit.periodicity,
                habit.creation_date),
            'consecutive_streaks': self.analytics_service.consecutive_streaks(
                habit.completion_logs,
                habit.periodicity,
                habit.creation_date)}



from typing import List, Optional

from sqlalchemy.orm import Session

from ...domain.model.habit import Habit
from ...domain.model.completion_log import CompletionLog
from ...domain.model.periodicity import Periodicity
from ...infrastructure.orm.models import CompletionLogModel, HabitModel


class HabitRepository:
    def __init__(self, session: Session):
        self.session = session

    def _to_domain_entity(self, habit_model: HabitModel) -> Habit:
        habit = Habit(
            id=habit_model.id,
            description=habit_model.description,
            periodicity=Periodicity(habit_model.periodicity),
            creation_date=habit_model.creation_date
        )
        habit.completion_logs = [
            CompletionLog(date=log.date)
            for log in habit_model.completion_logs
        ]
        return habit

    def _to_persistence_model(self, habit: Habit) -> HabitModel:
        habit_model = self.session.get(HabitModel, habit.id)
        habit_model.description = habit.description
        habit_model.periodicity = habit.periodicity.value
        habit_model.creation_date = habit.creation_date
        return habit_model

    def _to_new_persistence_model(self, habit: Habit) -> HabitModel:
        habit_model = HabitModel(
            description=habit.description,
            periodicity=habit.periodicity.value,
            creation_date=habit.creation_date
        )

        return habit_model

    def add(self, habit: Habit) -> Habit:
        habit_model = self._to_new_persistence_model(habit)
        self.session.add(habit_model)
        self.session.commit()
        return self._to_domain_entity(habit_model)

    def save(self, habit: Habit) -> None:
        habit_model = self._to_persistence_model(habit)
        existing_log_dates = {log.date for log in habit_model.completion_logs}
        for log in habit.completion_logs:
            if log.date not in existing_log_dates:
                habit_model.completion_logs.append(
                    CompletionLogModel(habit_id=habit.id, date=log.date)
                )
        self.session.commit()

    def remove(self, habit: Habit):
        db_habit = self.session.get(HabitModel, habit.id)
        if db_habit:
            self.session.delete(db_habit)
            self.session.commit()

    def find_by_id(self, habit_id: int) -> Optional[Habit]:
        habit_model = self.session.get(HabitModel, habit_id)
        return self._to_domain_entity(habit_model) if habit_model else None

    def get_all(self) -> List[Habit]:
        habit_models = self.session.query(HabitModel).all()
        return [self._to_domain_entity(habit_model)
                for habit_model in habit_models]

    def find_by_periodicity(self, periodicity: Periodicity) -> List[Habit]:
        habit_models = self.session.query(HabitModel).filter(
            HabitModel.periodicity == periodicity.value
        ).all()
        return [self._to_domain_entity(habit_model)
                for habit_model in habit_models]



from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from ..config import Config

engine = create_engine(Config.DATABASE_URL)

SessionLocal = scoped_session(sessionmaker(bind=engine))



import datetime
import random
from sqlalchemy.orm import sessionmaker

from .domain.model.periodicity import Periodicity
from .infrastructure.orm.models import Base, CompletionLogModel, HabitModel
from .infrastructure.orm.session import engine

Session = sessionmaker(bind=engine)


def generate_dates(start_date, periodicity, count, irregular=False):
    dates = []
    current_date = start_date

    for _ in range(count):
        if irregular and periodicity == Periodicity.DAILY:
            current_date += datetime.timedelta(days=random.randint(1, 2))
        elif periodicity == Periodicity.DAILY:
            current_date += datetime.timedelta(days=1)
        elif periodicity == Periodicity.WEEKLY:
            current_date += datetime.timedelta(weeks=1)
        elif periodicity == Periodicity.MONTHLY:
            current_date += datetime.timedelta(days=30)
        elif periodicity == Periodicity.YEARLY:
            current_date += datetime.timedelta(days=365)
        dates.append(current_date)

    return dates


def init_database():
    Base.metadata.drop_all(bind=engine)
    Base.metadata.create_all(bind=engine)

    session = Session()

    habits_data = [
        {"description": "60m Running",
         "periodicity": Periodicity.DAILY,
         "start_date": datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=29),
         "logs_count": 30},
        {"description": "10m Meditation",
         "periodicity": Periodicity.DAILY,
         "start_date": datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=29),
         "logs_count": 10,
         "irregular": True},
        {"description": "3h Hiking",
         "periodicity": Periodicity.WEEKLY,
         "start_date": datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(weeks=4),
         "logs_count": 4},
        {"description": "3h Swimming",
         "periodicity": Periodicity.MONTHLY,
         "start_date": datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=30 * 12),
         "logs_count": 12},
        {"description": "20d Traveling",
         "periodicity": Periodicity.YEARLY,
         "start_date": datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=365 * 3),
         "logs_count": 3}
    ]

    for habit_data in habits_data:
        habit = HabitModel(
            description=habit_data["description"],
            periodicity=habit_data["periodicity"],
            creation_date=habit_data["start_date"]
        )
        session.add(habit)
        log_dates = generate_dates(
            habit_data["start_date"],
            habit_data["periodicity"],
            habit_data["logs_count"],
            habit_data.get("irregular", False))

        for date in log_dates:
            log = CompletionLogModel(date=date, habit=habit)
            session.add(log)

    session.commit()
    session.close()

    return "Database initialized with sample data!"


if __name__ == "__main__":
    init_database()



from datetime import datetime
from typing import List

from ...domain.model.completion_log import CompletionLog
from ...domain.model.periodicity import Periodicity
from ...domain.service.streak_calculation_service import get_streaks_info
from ...infrastructure.repository.habit_repository import HabitRepository


class AnalyticsService:
    def __init__(self, repository: HabitRepository):
        self.repository = repository

    @staticmethod
    def current_streak(
            logs: List[CompletionLog],
            periodicity: str,
            creation_date: datetime,
            today: datetime = None) -> int:
        return get_streaks_info(
            logs,
            periodicity,
            creation_date,
            today).current_streak

    @staticmethod
    def longest_streak(
            logs: List[CompletionLog],
            periodicity: str,
            creation_date: datetime,
            today: datetime = None) -> int:
        return get_streaks_info(
            logs, periodicity, creation_date, today).max_streak

    @staticmethod
    def broken_streaks_count(
            logs: List[CompletionLog],
            periodicity: str,
            creation_date: datetime,
            today: datetime = None) -> int:
        return get_streaks_info(
            logs,
            periodicity,
            creation_date,
            today).broken_streaks

    @staticmethod
    def consecutive_streaks(
            logs: List[CompletionLog],
            periodicity: str,
            creation_date: datetime,
            today: datetime = None) -> int:
        return get_streaks_info(
            logs,
            periodicity,
            creation_date,
            today).consecutive_streaks

    @staticmethod
    def is_current_period_marked(
            logs: List[CompletionLog],
            periodicity: Periodicity,
            today: datetime) -> bool:
        if not logs:
            return False

        latest_log = max(logs, key=lambda log: log.date)
        start_of_next_period = periodicity.get_next_date(latest_log.date)

        return today < start_of_next_period

    def list_habits_by_periodicity(self, periodicity: Periodicity) -> list:
        return self.repository.find_by_periodicity(periodicity)

    def longest_streak_for_all_habits(self) -> int:
        habits = self.repository.get_all()
        max_streaks = [
            self.longest_streak(
                habit.completion_logs,
                habit.periodicity,
                habit.creation_date) for habit in habits]
        return max(max_streaks, default=0)

    def longest_streak_for_habit(self, habit_id: int) -> int:
        habit = self.repository.find_by_id(habit_id)
        return self.longest_streak(
            habit.completion_logs,
            habit.periodicity,
            habit.creation_date) if habit else 0

    def struggled_habits_last_month(self, today: datetime = None) -> list:

        habits = self.repository.get_all()
        struggled_habits = []

        for habit in habits:
            count = self.broken_streaks_count(
                habit.completion_logs,
                habit.periodicity,
                habit.creation_date,
                today)
            if count > 3:
                struggled_habits.append({"habit": {
                    "id": habit.id,
                    "description": habit.description,
                }})

        return struggled_habits



from fastapi import Depends
from sqlalchemy.orm import Session

from ..application.service.analytics_service import AnalyticsService
from ..application.service.habit_service import HabitService
from ..infrastructure.orm.session import SessionLocal
from ..infrastructure.repository.habit_repository import HabitRepository


def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


def get_habit_repository(db: Session = Depends(get_db)):
    return HabitRepository(db)


def get_analytics_service(
        habit_repository: HabitRepository = Depends(get_habit_repository)):
    return AnalyticsService(repository=habit_repository)


def get_habit_service(
        habit_repository: HabitRepository = Depends(get_habit_repository),
        analytics_service: AnalyticsService = Depends(get_analytics_service)):
    return HabitService(
        repository=habit_repository,
        analytics_service=analytics_service)



from datetime import datetime, timezone

from .completion_log import CompletionLog
from .periodicity import Periodicity


class Habit:
    def __init__(
            self,
            description: str,
            periodicity: Periodicity,
            creation_date: datetime = None,
            id: int = None):
        self.id = id
        self.description = description
        self.periodicity = periodicity
        self.creation_date = creation_date or datetime.now(timezone.utc)
        self.completion_logs = []

    def complete(self):
        completion_log = CompletionLog()
        self.completion_logs.append(completion_log)



from setuptools import find_packages, setup

setup(
    name='hTrackerIU',
    version='0.0.11',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    entry_points={
        'console_scripts': [
            'hTrackerIU=hTrackerIU.main:main',
        ],
    },
    install_requires=[
        'fastapi',
        'uvicorn',
        'sqlalchemy',
        'python-dateutil'
    ],
    extras_require={
        'dev': [
            'pytest',
            'httpx'
        ],
    },
    include_package_data=True,
    package_data={
        'hTrackerIU': [
            'static/*',
            'static/**/*',
            'static/**/**/*',
            'static/**/**/**/*'
        ],
    },
)



import argparse
import os
import webbrowser
from threading import Timer

import uvicorn
from sqlalchemy import create_engine, inspect

from .infrastructure.config import Config
from .init_db import init_database


def open_browser(port):
    webbrowser.open_new(f"http://localhost:{port}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, default=None,
                        help="Port to run the server on")
    args = parser.parse_args()

    port = args.port

    if port is None:
        port = int(os.getenv("PORT", 8095))

    engine = create_engine(Config.DATABASE_URL)
    inspector = inspect(engine)

    if not inspector.get_table_names():
        init_database()

    Timer(1, open_browser, args=(port,)).start()

    try:
        uvicorn.run(
            "hTrackerIU.api.main:app",
            host="0.0.0.0",
            port=port,
            reload=True)
    except SystemExit:
        print(f"ERROR: Port {port} is already in use.")
        print("You can specify a different port by using the --port argument")


if __name__ == "__main__":
    main()



from datetime import datetime, timezone

from sqlalchemy import Column, DateTime, Enum, ForeignKey, Integer, String
from sqlalchemy.orm import declarative_base, relationship

from ...domain.model.periodicity import Periodicity

Base = declarative_base()


class HabitModel(Base):
    __tablename__ = 'habits'

    id = Column(Integer, primary_key=True)
    description = Column(String)
    periodicity = Column(Enum(Periodicity))
    creation_date = Column(DateTime, default=datetime.now(timezone.utc))
    completion_logs = relationship(
        "CompletionLogModel", back_populates="habit")


class CompletionLogModel(Base):
    __tablename__ = 'completion_logs'

    id = Column(Integer, primary_key=True)
    date = Column(DateTime, default=datetime.now(timezone.utc))
    habit_id = Column(Integer, ForeignKey('habits.id'))
    habit = relationship("HabitModel", back_populates="completion_logs")



from datetime import datetime, timedelta
from enum import Enum

from dateutil.relativedelta import relativedelta


class Periodicity(Enum):
    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    MONTHLY = "MONTHLY"
    YEARLY = "YEARLY"

    def get_next_date(self, current_date: datetime) -> datetime:
        if self == Periodicity.DAILY:
            return current_date + timedelta(days=1)
        elif self == Periodicity.WEEKLY:
            return current_date + timedelta(weeks=1)
        elif self == Periodicity.MONTHLY:
            return (current_date + relativedelta(months=+1)).replace(day=1)
        elif self == Periodicity.YEARLY:
            return (
                current_date +
                relativedelta(
                    years=+
                    1)).replace(
                day=1,
                month=1)
        else:
            raise ValueError(f"Unsupported periodicity: {self.value}")
